/**
 *	@file
 *		AspinaCommands.cpp
 *	@brief
 *		Plexmotion Standard Link Interface Library
 *	@attention
 *		Copyright (C) 2021 - SHINANO-KENSHI Co,Ltd
 */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name, the copyright and the trademark of the SHINANO-KENSHI
 *     CO.,LTD. nor the names of its contributors may be used to endorse or
 *     promote products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "AspinaCommands.h" 

/**
 * @brief コンストラクタ
 */
AspinaCommands::AspinaCommands(void)
{
	for (int iLoop = 0; iLoop < 100; iLoop++) {
		buff[iLoop] = 0;
	}
	serialPub = 0;
	mReceived = false;
	serialMsg.data = "";

	for (int id = 0; id < ID_MAX; id++){
		mNDVersion[id] = 0;
		mNDAlarm[id]  = 0;
		mNDCurSpeed[id]  = 0;
		mNDCurTorque[id]  = 0;
		mNDStatus[id]  = 0;
		mNDServoMode[id]  = 0;
		mNDOrderMode[id][0]  = 0;
		mNDOrderMode[id][1]  = 0;
		mNDOrderMode[id][2]  = 0;
		mNDOrderMode[id][3]  = 0;
		mNDOrderTorque[id]  = 0;
		mNDOrderSpeed[id]  = 0;
		mNDLimitSpeed[id]  = 0;
		mNDLimitCurrent[id]  = 0;
		mNDServoOffMode[id]  = 0;
		mNDSystemLimitSpeed[id]  = 0;
		mNDSystemLimitCurrent[id]  = 0;
	}
}

/**
 * @brief アスキー文字を16進数へ変換
 * @param [in] 変換したい文字
 * @return 16進数値
 */
int AspinaCommands::asciiToHex(char asciiChar)
{
	int16_t iRet;

	asciiChar = tolower(asciiChar);
	
	if(asciiChar >= 'a'){
		iRet = asciiChar - 'a' + 10;
	}
	else{
		iRet = asciiChar - '0';
	}
	
	return iRet;
}

/**
 * @brief 配信ハンドルの登録
 * @param [in] set publisherのポインタ
 */
void AspinaCommands::SetPublisherHandle(ros::Publisher* set)
{
    serialPub = set;
}

/**
 * @brief 受信フラグクリア
 */
void AspinaCommands::ClearReceiveFlag(void)
{
	mReceived = false;
}

int16_t GetInt16(const uint8_t* pbyData)
{
	int16_t wData;
	wData = (int16_t)((int16_t)pbyData[0] << 8);
	wData += ((int16_t)pbyData[1]);
	return wData;
}

int32_t GetInt32(const uint8_t* pbyData)
{
	int32_t dwData;
	dwData = ((int32_t)pbyData[0] << 24);
	dwData += ((int32_t)pbyData[1] << 16);
	dwData += ((int32_t)pbyData[2] << 8);
	dwData += ((int32_t)pbyData[3]);
	return dwData;
}

/**
 * @brief 購読時のコールバック
 * @param [in] 受信文字列
 */
void AspinaCommands::Callback(const char* str)
{
	uint8_t msgSize = strlen(str);
	const uint8_t ASCII_SIZE = 40;
	char asciiBuff[ASCII_SIZE] = {0};
	
	if (msgSize > ASCII_SIZE) {
		ROS_ERROR("ascii size over %d\n", msgSize);
		return;
	}

	for (uint8_t i = 0; i < msgSize; i++) {
		if( str[i] >='a'){
			asciiBuff[i] = str[i] - 'a' + 10;
		}
		else{
			asciiBuff[i] = str[i] - '0';
		}
   	 }
    
	char hexBuff[(msgSize /= 2)];
	for (int j = 0; j < msgSize; j++) {
		hexBuff[j] = asciiBuff[(j * 2)] * 16 + asciiBuff[(j * 2 + 1)];
	}
	
	/* format
	0 … 通信id
	1 … ステータス
	2 … データ長
	3 … パラメータID
	4 以降 データ
	最終バイト … チェックサム */
	
	switch (hexBuff[3]) {	// 3バイト目がパラメータID
	case ASPINA_PID_VERSION:
		mNDVersion[hexBuff[0]] = GetInt16((uint8_t*)&hexBuff[4]);
		break;
    	
	case ASPINA_PID_CURRENT_TORQUE:
		mNDCurTorque[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[4]);
		break;
		
	case ASPINA_PID_CURRENT_SPEED:
		mNDCurSpeed[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[4]);
		break;
		
	case ASPINA_PID_CURRENT_ALARM:
		mNDAlarm[hexBuff[0]] = hexBuff[4];
		break;
		
	case ASPINA_PID_STATUS:
		mNDStatus[hexBuff[0]] = GetInt16((uint8_t*)&hexBuff[4]);
		break;
		
	case ASPINA_PID_SERVOOFF_MODE:
		mNDServoOffMode[hexBuff[0]] = hexBuff[4];
		break;

	case ASPINA_PID_SERVO_MODE:
		mNDServoMode[hexBuff[0]] = hexBuff[4];
		break;
		
	case ASPINA_PID_ORDER_MODE:
		if((hexBuff[4] >= 1) && (hexBuff[4] <= 3)){
			mNDOrderMode[hexBuff[0]][hexBuff[4]] = hexBuff[5];
		}
		break;
		
	case ASPINA_PID_ORDER_TORQUE:
		mNDOrderTorque[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[4]);
		mNDOrderSpeed[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[8]);
		break;
		
	case ASPINA_PID_ORDER_SPEED:
		mNDOrderSpeed[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[4]);
		break;
		
	case ASPINA_PID_LIMITTER:
		mNDLimitCurrent[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[4]);
		mNDLimitSpeed[hexBuff[0]] = GetInt32((uint8_t*)&hexBuff[8]);
		break;

	default:
		break;
	}
	mReceived = true;
}

/**
 * @brief ファームバージョンを取得
 * @param [in] id 通信ID
 * @return ファームバージョン
 */
uint16_t AspinaCommands::GetVersion(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDVersion[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 現在アラームを取得
 * @param [in] id 通信ID
 * @return アラーム番号
 */
uint16_t AspinaCommands::GetAlarm(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDAlarm[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 現在速度を取得
 * @param [in] id 通信ID
 * @return 現在速度[pps * 10]
 */
int32_t AspinaCommands::GetCurSpeed(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDCurSpeed[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 指令電流を取得
 * @param [in] id 通信ID
 * @return 指令電流[mA]
 */
int32_t AspinaCommands::GetCurTorque(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDCurTorque[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief サーボオンオフ状態を取得
 * @param [in] id 通信ID
 * retval 0 サーボオフ
 * retval 1 サーボオン
 */
uint16_t AspinaCommands::IsServoOn(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return (mNDStatus[id] & 0x02)>>1;
	}
	else{
		return 0;
	}
}

/**
 * @brief サーボモードを取得
 * @param [in] id 通信ID
 * @retval 1 トルク制御
 * @retval 2 速度制御
 * @retval 3 位置制御
 */
uint16_t AspinaCommands::GetServoMode(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDServoMode[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 指令方法を取得
 * @param [in] servoMode モード指定
 * @param [in] id 通信ID
 * @retval 0 通信
 * @retval 1 アナログ
 */
uint16_t AspinaCommands::GetOrderMode(uint8_t servoMode, uint8_t id = 1)
{
	if(id <= ID_MAX){
		if((0 < servoMode) && (servoMode < 4)){
			return mNDOrderMode[id][servoMode];
		}
		else{
			return 0;
		}
	}
	else{
		return 0;
	}
}

/**
 * @brief 速度指令方法を取得
 * @param [in] id 通信ID
 * @retval 0 通信
 * @retval 1 アナログ
 */
uint16_t AspinaCommands::GetOrderSpeedMode(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return GetOrderMode(ASPINA_SERVOMODE_SPEED, id);
	}
	else{
		return 0;
	}
}

/**
 * @brief 指令電流を取得
 * @param [in] id 通信ID
 * @return 指令電流[mA]
 */
int32_t AspinaCommands::GetOrderTorque(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDOrderTorque[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 指令速度を取得
 * @param [in] id 通信ID
 * @return 指令速度[pps * 10]
 */
int32_t AspinaCommands::GetOrderSpeed(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDOrderSpeed[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 速度リミットを取得
 * @param [in] id 通信ID
 * @return 速度リミット[pps * 10]
 */
int32_t AspinaCommands::GetLimitSpeed(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDLimitSpeed[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief サーボオフモードを取得
 * @param [in] id 通信ID
 * @retval 0 フリー
 * @retval 1 ショートブレーキ
 */
uint16_t AspinaCommands::GetServoOffMode(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDServoOffMode[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 電流リミットを取得
 * @param [in] id 通信ID
 * @return 電流リミット[mA]
 */
int32_t AspinaCommands::GetLimitCurrent(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDLimitCurrent[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief システム速度リミットを取得
 * @param [in] id 通信ID
 * @return システム速度リミット[pps * 10]
 */
int32_t AspinaCommands::GetSytemLimitSpeed(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDSystemLimitSpeed[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief システム電流リミットを取得
 * @param [in] id 通信ID
 * @return システム電流リミット[mA]
 */
int32_t AspinaCommands::GetSytemLimitCurrent(uint8_t id = 1)
{
	if(id <= ID_MAX){
		return mNDSystemLimitCurrent[id];
	}
	else{
		return 0;
	}
}

/**
 * @brief 指令トピックコマンド文字列作成
 * @return コマンド文字列
 */
char* AspinaCommands::CreateCommand(const char* str, uint8_t id = 1)
{
	sprintf(&buff[0], str, id);
	return buff;
}

/**
 * @brief 指令トピックコマンド文字列作成、引数1つ
 * @return コマンド文字列
 */
char* AspinaCommands::CreateCommand(const char* str, int32_t x, uint8_t id = 1)
{
	sprintf(&buff[0], str, id, x);
	return buff;
}

/**
 * @brief 指令トピックコマンド文字列作成、引数2つ
 * @return コマンド文字列
 */
char* AspinaCommands::CreateCommand(const char* str, int32_t x, int32_t y, uint8_t id = 1)
{
	sprintf(&buff[0], str, id, x, y);
	return buff;
}

/**
 * @brief 文字列をパブリッシュ
 */
void AspinaCommands::PublishString(const char* str, uint8_t id = 1)
{
	mReceived = false;
	serialMsg.data = CreateCommand(str, id);
	serialPub->publish(serialMsg);
	WaitReceive();
}

/**
 * @brief コマンドトピックを発行
 */
void AspinaCommands::PublishCommand(const char* str, uint8_t id = 1)
{
	PublishString(CreateCommand(str, id), id);
}

/**
 * @brief コマンドトピックを発行、引数1つ
 */
void AspinaCommands::PublishCommand(const char* str, int32_t x, uint8_t id = 1)
{
	PublishString(CreateCommand(str, x, id), id);
}

/**
 * @brief コマンドトピックを発行、引数2つ
 * @param str コマンド文字列, id 通信ID , x  引数1 y 引数2 
 */
void AspinaCommands::PublishCommand(const char* str, int32_t x, int32_t y, uint8_t id = 1)
{
	PublishString(CreateCommand(str, x, y, id), id);
}

/**
 * @brief 受信フラグを返す
 * @retval 0 未受信
 * @retval 1 受信
 */
bool AspinaCommands::IsReceived(void)
{
	return mReceived;
}

/**
 * @brief 受信フラグを待つ
 */
void AspinaCommands::WaitReceive(void)
{
	ros::Rate loop_rate(1000);
	do {
		ros::spinOnce();
    		loop_rate.sleep();
	} while((ros::ok()) && (mReceived == false));
}


/**
 * @brief バージョン取得にてハンドシェイクを行う
 * @param [in] id 通信ID
 */
void AspinaCommands::Handshake(uint8_t id = 1)
{
	ROS_INFO("Get Version");
	ros::Rate rate(100);
	do {
		serialMsg.data = CreateCommand(ASPINA_GET_VERSION, id);
		serialPub->publish(serialMsg);
		ros::spinOnce();
		rate.sleep();
	} while ((ros::ok()) && (mReceived == false));
	
	//指令モードを通信モードに設定
	PublishCommand(ASPINA_SET_ORDER_TORQUE_MODE, ASPINA_ORDERMODE_DIRECT, id);
	PublishCommand(ASPINA_SET_ORDER_SPEED_MODE, ASPINA_ORDERMODE_DIRECT, id);
	
	ROS_INFO("id : %d, version : %d", id, GetVersion(id));	// バージョン取得完了
}
