/**
 *	@file
 *		AspinaMotorDrive.cpp
 *	@brief
 *		Plexmotion Standard Link Motor Control Node
 *	@attention
 *		Copyright (C) 2021 - SHINANO-KENSHI Co,Ltd
 */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name, the copyright and the trademark of the SHINANO-KENSHI
 *     CO.,LTD. nor the names of its contributors may be used to endorse or
 *     promote products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <fcntl.h>
#include <termios.h>
#include <ros/ros.h>
#include <std_msgs/String.h>
#include "param.h"
/*----------------------------------------------------------------------------------------------*/
/*定義*/
/*----------------------------------------------------------------------------------------------*/
#define PACKETBUFFSIZE	100

typedef enum ENUM_STDLINK_PACKET_STATE {
	STDLINK_PACKET_ID = 0,
	STDLINK_PACKET_CMDSTS = 1,
	STDLINK_PACKET_LENGTH = 2,
	STDLINK_PACKET_DATA = 3,
	STDLINK_PACKET_CHECKSUM = 4
} EN_STDLINK_PACKET_STATE;

typedef struct STRUCT_STD_PACKET {
	EN_STDLINK_PACKET_STATE eState;
	
	unsigned char byID;
	unsigned char byCheckSum;
	unsigned char byCount;
	
	unsigned char byBuffSize;
	unsigned char* pbyBuff;
} ST_STDLINK_PACKET;


/*----------------------------------------------------------------------------------------------*/
/*ファイル内共通変数*/
/*----------------------------------------------------------------------------------------------*/
int16_t fd1;
static unsigned char m_abyPacketBuff[PACKETBUFFSIZE];
static bool m_bWaitCommand = false;

/*----------------------------------------------------------------------------------------------*/
/*ファイル内関数*/
/*----------------------------------------------------------------------------------------------*/
void StdLinkPacket_Init(ST_STDLINK_PACKET* pstSelf);
bool StdLinkPacket_CreateStatic(ST_STDLINK_PACKET* pstSelf, unsigned char byBuffSize, unsigned char* pbyBuff);
bool StdLinkPacket_SetID(ST_STDLINK_PACKET* pstSelf, unsigned char byID);
unsigned char StdLinkPacket_GetPacketSize(ST_STDLINK_PACKET* pstSelf);
bool StdLinkPacket_Check(ST_STDLINK_PACKET* pstSelf, unsigned char byData);


/**
 * @brief シリアルポート初期化
 * @param [in] deviceName オープンするシリアルデバイスの名前
 * @retval 0 ポートオープン失敗
 * @retval 1 ポートオープン成功 
 */
bool OpenSerial(const char *deviceName)
{
	fd1 = open(deviceName, O_RDWR | O_NONBLOCK);	// non blocking I/O
	if (fd1 < 0){
		return false;
	}

	struct termios conf_tio;
	memset(&conf_tio, 0, sizeof(conf_tio));
	conf_tio.c_cflag = CS8 | CLOCAL | CREAD ;// Start=1, Stop=1, Data=8Bits
	conf_tio.c_cc[VTIME] = 0;
	conf_tio.c_cc[VMIN] = 0;
	speed_t BAUDRATE = B115200;
	cfsetispeed(&conf_tio, BAUDRATE);
	cfsetospeed(&conf_tio, BAUDRATE);
	tcsetattr(fd1, TCSANOW, &conf_tio);

	ROS_INFO("Serial Open Succeed");
	return true;
}

/**
 * @brief アスキーデータ2文字1バイトの文字列を数値列に変換
 * @param [in] pIn 入力文字列
 * @param [out] pOut 出力データ列
 * @param [in] pIn 入力文字列のサイズ
 */
void AsciiBuffToByteBuff(const char *pIn, char *pOut, uint8_t iSize)
{
	char buff[PACKETBUFFSIZE];

	for (uint8_t i = 0; i < iSize; i++) {
		if(pIn[i] >='a'){
			buff[i] = pIn[i] - 'a' + 10;
		}
		else{
			buff[i] = pIn[i] - '0';
		}
	}

	for (uint8_t i = 0; i < (iSize / 2); i++) {
		uint8_t hex = 0;
		hex = buff[(i * 2)] << 4;
		hex &= 0xF0;
		hex += buff[(i * 2 + 1)];
        
		pOut[i] = hex;
	}
}

/**
 * @brief チェックサムの計算
 * @param [in] 入力データ列
 * @return チェックサム計算値
 */
char AddCheckSum(char *pIn)
{
	unsigned int calcSum = 0;
	char cPos;
	for (cPos = 0; cPos < (pIn[2] + 3); cPos++) {
		calcSum += pIn[cPos];
	}
	calcSum &= (unsigned int)0x00ff;
	pIn[cPos] = 0xff - (unsigned char)calcSum;
	cPos += 1;

	return cPos;
}

/**
 * @brief ドライバへのコマンド送信
 * @param [in] pIn ドライバへの送信データ列
 */
void SendMessegeToDriver(const char *pIn)
{
	char hexBuff[PACKETBUFFSIZE] = {0};
	char buff[PACKETBUFFSIZE] = "8";
	strcat(buff, pIn);

	AsciiBuffToByteBuff(buff, hexBuff, strlen(buff));
	char buffLen = AddCheckSum(hexBuff);
       
	write(fd1, hexBuff, buffLen);
}

/**
 * @brief 購読時のコールバック
 * @param [in] serialMsg 購読したメッセージ
 */
 void SerialCallback(const std_msgs::String& serialMsg)
{
	uint8_t msgSize = serialMsg.data.size();
	const uint8_t ASCII_SIZE = 40;
    
	if (msgSize > ASCII_SIZE) {
		ROS_ERROR("ascii size over %d\n", msgSize);
		return;
	}

	SendMessegeToDriver((const char *)(&serialMsg.data[0]));
	m_bWaitCommand = true;
}

/**
 * @brief ドライバからの応答コマンド受信
 * @param [out] pRecvBuff 1パケット受信完了時、パケットを保存する
 * @return 受信パケットサイズ
 */
int16_t ReceiveOnePacket(unsigned char *pRecvBuff)
{
	ST_STDLINK_PACKET stPacket;

	StdLinkPacket_CreateStatic(&stPacket, PACKETBUFFSIZE, pRecvBuff);

	ros::Rate loop_rate(1000);
    
	uint16_t iTimeOut = 0;
	bool bWait = false;
	do {
		char buf;
		while (read(fd1, &buf, 1) == 1){
			bool bCheck = StdLinkPacket_Check(&stPacket, buf);
      
			if (bCheck == true) {
				bWait = true;
			}
		}
		ros::spinOnce();
		loop_rate.sleep();
		iTimeOut++;
	} while ((ros::ok()) && (bWait == false) && (iTimeOut < 500));

	m_bWaitCommand = false;

	int16_t iRetSize;
	if (iTimeOut < 500) {
		iRetSize = StdLinkPacket_GetPacketSize(&stPacket);
	}
	else {
		iRetSize = -1;
	}
    return iRetSize;
}


/**
 * @brief メイン関数
 * @param [in] argc 引数の数
 * @param [out] argv 引数文字列
 */
int main(int argc, char **argv)
{
	ROS_INFO("COM : %s", argv[1]);
	ros::init(argc, argv, "driveMotor");
	ros::NodeHandle n;

	std::string pubMessageName = "AspinaResponse";
	std::string subMessageName = "AspinaOrder";
    

	
	ros::Publisher serialPub = n.advertise<std_msgs::String>(pubMessageName, 1000);
	ros::Subscriber serialSub = n.subscribe(subMessageName, 20, SerialCallback);

	ros::NodeHandle pnh("~");
	std::string serialName = "/dev/ttyUSB0";
	pnh.getParam("com", serialName);
	bool res = OpenSerial((char*)serialName.c_str());
	if(res == false){
		ROS_ERROR("Serial Fail: cound not open %s", serialName.c_str());
		ROS_INFO("Serial Fail");
		ros::shutdown();
	}



	ros::Rate loop_rate(1000);
	while (ros::ok()){
		if (m_bWaitCommand == true) {
			std_msgs::String serialMsg;
		int16_t recvDataSize = ReceiveOnePacket((unsigned char*)m_abyPacketBuff);
			if (recvDataSize == -1) {
				//タイムアウト
			}
			else {
				char sendBuff[256] = {0};
				char tempBuff[4] = {0};
				for(uint16_t i = 0; i < recvDataSize; i++){
					sprintf(tempBuff, "%02x", m_abyPacketBuff[i]);
					strcat(sendBuff, tempBuff);
				}
				serialMsg.data = sendBuff;
				serialPub.publish(serialMsg);
			}
		}
        ros::spinOnce();
        loop_rate.sleep();
    }
    
    return 0;
}

/**
 * @brief パケット構造解析の初期化
 * @param [in] pstSelf パケット構造解析用構造体
 */
void StdLinkPacket_Init(ST_STDLINK_PACKET* pstSelf)
{
	unsigned char i;

	pstSelf->eState = STDLINK_PACKET_ID;
	pstSelf->byCheckSum = 0;
	pstSelf->byCount = 0;

	for(i = 0; i < pstSelf->byBuffSize; i++){
		pstSelf->pbyBuff[i] = 0;
	}
}

/**
 * @brief パケット構造解析のメモリ初期化
 * @param [in] pstSelf パケット構造解析用構造体
 * @param [in] byBuffSize バッファサイズ
 * @param [in] pbyBuff バッファの実体
 * @retval true 初期化成功
 * @retval false 初期化失敗
 */
bool StdLinkPacket_CreateStatic(ST_STDLINK_PACKET* pstSelf, unsigned char byBuffSize, unsigned char* pbyBuff)
{
	if (byBuffSize < 4) {
		return false;
	}

	pstSelf->byBuffSize = byBuffSize;
	pstSelf->pbyBuff = pbyBuff;

	StdLinkPacket_Init(pstSelf);

	return true;
}

/**
 * @brief 通信ID設定
 * @param [in] pstSelf パケット構造解析用構造体
 * @param [in] byID 対象のドライバID
 * @retval true 完了
 */
bool StdLinkPacket_SetID(ST_STDLINK_PACKET* pstSelf, unsigned char byID)
{
	pstSelf->byID = byID;
	return true;
}

/**
 * @brief 受信データサイズ取得
 * @param [in] pstSelf パケット構造解析用構造体
 * @return データサイズ
 */
unsigned char StdLinkPacket_GetPacketSize(ST_STDLINK_PACKET* pstSelf)
{
	return (unsigned char)(pstSelf->pbyBuff[2] + 4);
}


/**
 * @brief 受信データプロトコル解析
 * @param [in] pstSelf パケット構造解析用構造体
 * @param [in] byData パケットに追加するデータ
 * @retval true 正常パケット受信
 * @retval false パケット未完
 */
bool StdLinkPacket_Check(ST_STDLINK_PACKET* pstSelf, unsigned char byData)
{
	bool bResult = false;

	switch(pstSelf->eState){
	case STDLINK_PACKET_ID:
		if ((byData <= ID_MAX) ||
			(byData == 0xFB)) {
			pstSelf->pbyBuff[0] = byData;
			pstSelf->byCheckSum = byData;
			pstSelf->eState = STDLINK_PACKET_CMDSTS;
		}
		break;

	case STDLINK_PACKET_CMDSTS:
		if (byData <= 0x04) {
			pstSelf->pbyBuff[1] = byData;
			pstSelf->byCheckSum += byData;
			pstSelf->eState = STDLINK_PACKET_LENGTH;
		}
		else {
			pstSelf->eState = STDLINK_PACKET_ID;
		}
		break;

	case STDLINK_PACKET_LENGTH:
		pstSelf->pbyBuff[2] = byData;
		pstSelf->byCheckSum += byData;
		if (byData == 0) {
			pstSelf->byCount = 0;
			pstSelf->eState = STDLINK_PACKET_CHECKSUM;
		}
		else if (byData < (pstSelf->byBuffSize - (unsigned char)3)) {
			pstSelf->byCount = 0;
			pstSelf->eState = STDLINK_PACKET_DATA;
		}
		else {
			pstSelf->eState = STDLINK_PACKET_ID;
		}
		break;

	case STDLINK_PACKET_DATA:
		pstSelf->pbyBuff[pstSelf->byCount + 3] = byData;
		pstSelf->byCheckSum += byData;
		pstSelf->byCount++;
		if (pstSelf->byCount >= pstSelf->pbyBuff[2]) {
			pstSelf->eState = STDLINK_PACKET_CHECKSUM;
		}
		break;

	case STDLINK_PACKET_CHECKSUM:
		pstSelf->pbyBuff[pstSelf->byCount + 3] = byData;
		pstSelf->byCheckSum += byData;
		pstSelf->byCheckSum = (unsigned char)((unsigned char)0xFF - pstSelf->byCheckSum);

		if (pstSelf->byCheckSum == 0) {
			bResult = true;
		}

		pstSelf->eState = STDLINK_PACKET_ID;
		break;

	default:
		pstSelf->eState = STDLINK_PACKET_ID;
		break;
	}

	return bResult;
}

