/**
 *	@file
 *		UserSample.cpp
 *	@brief
 *		Plexmotion Standard Link Test Node
 *	@attention
 *		Copyright (C) 2021 - SHINANO-KENSHI Co,Ltd
 */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name, the copyright and the trademark of the SHINANO-KENSHI
 *     CO.,LTD. nor the names of its contributors may be used to endorse or
 *     promote products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <ros/ros.h> 
#include "AspinaCommands.h"

AspinaCommands Aspina;	/** Aspinaノードを扱うためのクラス定義 */

/**
 * @brief 購読時のコールバック
 * @param [in] 購読したメッセージ
 */
void SubscribeCallback(const std_msgs::String& responseMsg)
{
	/**
	 * AspinaMotorDriveノードから発行されたメッセージを
	 * AspinaクラスのCallback関数へ入力することにより、データの読み取りが容易になります。
	 */
	Aspina.Callback(responseMsg.data.c_str());
}

/** 
 * @brief メイン関数
 */
int main(int argc, char **argv)
{
	// ROSノードとしての初期化
	ros::init(argc, argv, "userSample");
	
	/* ノードハンドラ作成 */
	ros::NodeHandle n;
 
	/* パブリッシャ・サブスクライバ作成 */
	ros::Publisher serialPub = n.advertise<std_msgs::String>("AspinaOrder", 1000);
	ros::Subscriber responseSub = n.subscribe("AspinaResponse", 20, SubscribeCallback);
    
	/* AspinaCommandsへpublisherのポインタを設定 */
	Aspina.SetPublisherHandle(&serialPub);
 
	/* モータ初期設定 */
	uint8_t id = 1;
	Aspina.Handshake(id);

	ROS_INFO("ServoOff");
	Aspina.PublishCommand(ASPINA_SET_SERVO_ONOFF, ASPINA_SERVO_OFF, id);

	ROS_INFO("SetServoMode");
	Aspina.PublishCommand(ASPINA_SET_SERVO_MODE, ASPINA_SERVOMODE_SPEED, id);

	ROS_INFO("OrderSpeed");
	Aspina.PublishCommand(ASPINA_SET_ORDER_SPEED, 2000, id);

	ROS_INFO("ServoOn");
	Aspina.PublishCommand(ASPINA_SET_SERVO_ONOFF, ASPINA_SERVO_ON, id);

	// 回転速度の設定と監視
	for (uint8_t iLoop = 0; iLoop < 100; iLoop++) {
		int32_t orderSpeed = 2000 + iLoop * 10;
		Aspina.PublishCommand(ASPINA_SET_ORDER_SPEED, orderSpeed, id);
		Aspina.PublishCommand(ASPINA_GET_CURRENT_SPEED, id);
		ROS_INFO("Speed %d", Aspina.GetCurSpeed(id));
	}
	
	// モータ停止
	ROS_INFO("ServoOff");
	Aspina.PublishCommand(ASPINA_SET_SERVO_ONOFF, ASPINA_SERVO_OFF, id);
	
	return 0;
}
